# -*- coding: utf-8 -*-
"""
   click_default_group
   ~~~~~~~~~~~~~~~~~~~

   Site: https://github.com/sublee/click-default-group
   License: BSD 3 clause
     https://github.com/sublee/click-default-group/blob/master/LICENSE

   Define a default subcommand by `default=True`:
   .. sourcecode:: python
      import click
      from click_default_group import DefaultGroup
      @click.group(cls=DefaultGroup, default_if_no_args=True)
      def cli():
          pass
      @cli.command(default=True)
      def foo():
          click.echo('foo')
      @cli.command()
      def bar():
          click.echo('bar')
   Then you can invoke that without explicit subcommand name:
   .. sourcecode:: console
      $ cli.py --help
      Usage: cli.py [OPTIONS] COMMAND [ARGS]...
      Options:
        --help    Show this message and exit.
      Command:
        foo*
        bar
      $ cli.py
      foo
      $ cli.py foo
      foo
      $ cli.py bar
      bar
"""
import click

__all__ = ['DefaultGroup']
__version__ = '1.1'


class DefaultGroup(click.Group):
    """Invokes a subcommand marked with `default=True` if any subcommand not
    chosen.
    :param default_if_no_args: resolves to the default command if no arguments
                               passed.
    """

    def __init__(self, *args, **kwargs):
        self.default_cmd_name = None
        self.default_if_no_args = kwargs.pop('default_if_no_args', False)
        # To resolve as the default command.
        if not kwargs.get('ignore_unknown_options', True):
            raise ValueError('Default group accepts unknown options')
        self.ignore_unknown_options = True
        super(DefaultGroup, self).__init__(*args, **kwargs)

    def command(self, *args, **kwargs):
        default = kwargs.pop('default', False)
        decorator = super(DefaultGroup, self).command(*args, **kwargs)
        if not default:
            # Customized feature not used.
            return decorator

        def _decorator(f):
            cmd = decorator(f)
            if default:
                if self.default_cmd_name is not None:
                    del self.commands[cmd.name]
                    raise RuntimeError('Default command already defined')
                self.default_cmd_name = cmd.name
            return cmd

        return _decorator

    def parse_args(self, ctx, args):
        if not args and self.default_if_no_args:
            args.insert(0, self.default_cmd_name)
        return super(DefaultGroup, self).parse_args(ctx, args)

    def get_command(self, ctx, cmd_name):
        if cmd_name not in self.commands:
            # No command name matched.
            ctx.arg0 = cmd_name
            cmd_name = self.default_cmd_name
        return super(DefaultGroup, self).get_command(ctx, cmd_name)

    def resolve_command(self, ctx, args):
        base = super(DefaultGroup, self)
        cmd_name, cmd, args = base.resolve_command(ctx, args)
        if hasattr(ctx, 'arg0'):
            args.insert(0, ctx.arg0)
        return cmd_name, cmd, args

    def format_commands(self, ctx, formatter):
        formatter = DefaultCommandFormatter(self, formatter, mark='*')
        return super(DefaultGroup, self).format_commands(ctx, formatter)


class DefaultCommandFormatter(object):
    """Wraps a formatter to mark a default command."""

    def __init__(self, group, formatter, mark='*'):
        self.group = group
        self.formatter = formatter
        self.mark = mark

    def __getattr__(self, attr):
        return getattr(self.formatter, attr)

    def write_dl(self, rows, *args, **kwargs):
        rows_ = []
        for cmd_name, help in rows:
            if cmd_name == self.group.default_cmd_name:
                rows_.insert(0, (cmd_name + self.mark, help))
            else:
                rows_.append((cmd_name, help))
        return self.formatter.write_dl(rows_, *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Manage multiple git, mercurial, svn repositories from a YAML / JSON file.

vcspull
~~~~~~~

:copyright: Copyright 2013-2018 Tony Narlock.
:license: MIT, see LICENSE for details

"""

from __future__ import absolute_import, print_function, unicode_literals

# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

from . import cli  # NOQA

logging.getLogger(__name__).addHandler(NullHandler())



# -*- coding: utf-8 -*-
"""Utility functions for vcspull.

vcspull.util
~~~~~~~~~~~~

"""
from __future__ import absolute_import, print_function, unicode_literals

import os

from libvcs._compat import PY2

if PY2:
    from collections import Mapping
else:
    from collections.abc import Mapping

CONFIG_DIR = os.path.expanduser('~/.vcspull/')  # remove dupes of this


def update_dict(d, u):
    """Return updated dict.

    http://stackoverflow.com/a/3233356

    :param d: dict
    :type d: dict
    :param u: updated dict.
    :type u: dict
    :rtype: dict

    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = update_dict(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d



# -*- coding: utf-8 -*-
"""CLI utilities for vcspull.

vcspull.cli
~~~~~~~~~~~

"""
from __future__ import absolute_import, print_function

import logging
import sys
from copy import deepcopy

import click

from libvcs._compat import PY2
from libvcs.shortcuts import create_repo_from_pip_url

from .__about__ import __version__
from .cli_defaultgroup import DefaultGroup
from .config import filter_repos, find_config_files, load_configs
from .log import DebugLogFormatter, RepoFilter, RepoLogFormatter

if PY2:
    FileNotFoundError = EnvironmentError

MIN_ASYNC = 3  # minimum amount of repos to sync concurrently
MAX_ASYNC = 8  # maximum processes to open:w

log = logging.getLogger(__name__)


def setup_logger(log=None, level='INFO'):
    """Setup logging for CLI use.

    :param log: instance of logger
    :type log: :py:class:`Logger`

    """
    if not log:
        log = logging.getLogger()
    if not log.handlers:
        channel = logging.StreamHandler()
        channel.setFormatter(DebugLogFormatter())

        log.setLevel(level)
        log.addHandler(channel)

        # setup styling for repo loggers
        repo_logger = logging.getLogger('libvcs')
        channel = logging.StreamHandler()
        channel.setFormatter(RepoLogFormatter())
        channel.addFilter(RepoFilter())
        repo_logger.setLevel(level)
        repo_logger.addHandler(channel)


@click.group(cls=DefaultGroup, default_if_no_args=True)
@click.option(
    '--log-level',
    default='INFO',
    help='Log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)',
)
@click.version_option(version=__version__, message='%(prog)s %(version)s')
def cli(log_level):
    setup_logger(log=log, level=log_level.upper())


@cli.command(name='update', default=True)
@click.argument('repo_terms', nargs=-1)
@click.option(
    '--run-async',
    '-a',
    is_flag=True,
    help='Run repo syncing concurrently (experimental)',
)
@click.option(
    '--log-level',
    default='INFO',
    help='Log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)',
)
@click.option('config', '-c', type=click.Path(exists=True), help='Specify config')
def update(repo_terms, run_async, log_level, config):
    setup_logger(log=log, level=log_level.upper())

    if config:
        configs = load_configs([config])
    else:
        configs = load_configs(find_config_files(include_home=True))
    found_repos = []

    if repo_terms:
        for repo_term in repo_terms:
            repo_dir, vcs_url, name = None, None, None
            if any(repo_term.startswith(n) for n in ['./', '/', '~', '$HOME']):
                repo_dir = repo_term
            elif any(repo_term.startswith(n) for n in ['http', 'git', 'svn', 'hg']):
                vcs_url = repo_term
            else:
                name = repo_term

            # collect the repos from the config files
            found_repos.extend(
                filter_repos(configs, repo_dir=repo_dir, vcs_url=vcs_url, name=name)
            )
    else:
        found_repos = configs

    found_repos_n = len(found_repos)
    # turn them into :class:`Repo` objects and clone/update them
    if run_async and found_repos_n >= MIN_ASYNC:
        from multiprocessing import Pool

        p = Pool(clamp(found_repos_n, MIN_ASYNC, MAX_ASYNC))
        p.map_async(update_repo, found_repos).get()
    else:
        list(map(update_repo, found_repos))


def clamp(n, _min, _max):
    return max(_min, min(n, _max))


def progress_cb(output, timestamp):
    sys.stdout.write(output)
    sys.stdout.flush()


def update_repo(repo_dict):
    repo_dict = deepcopy(repo_dict)
    if 'pip_url' not in repo_dict:
        repo_dict['pip_url'] = repo_dict.pop('url')
    repo_dict['progress_callback'] = progress_cb

    r = create_repo_from_pip_url(**repo_dict)  # Creates the repo object

    remote_settings = repo_dict.get('remotes', [])
    if not any(rs for rs in remote_settings if rs['remote_name'] == 'origin'):
        remote_settings.append({'remote_name': 'origin', 'url': repo_dict['pip_url']})

    for remote_setting in remote_settings:
        config_remote_name = remote_setting['remote_name']  # From config file
        try:
            current_remote = r.remotes_get.get(config_remote_name)
        except FileNotFoundError:  # git repo doesn't exist yet, so cna't be outdated
            break

        if current_remote is not None and current_remote[0] != remote_setting['url']:
            print(
                'Updating remote {name} ({current_url}) with {new_url}'.format(
                    name=config_remote_name,
                    current_url=current_remote[0],
                    new_url=remote_setting['url'],
                )
            )
            r.remote_set(
                name=config_remote_name, url=remote_setting['url'], overwrite=True
            )
    r.update_repo()  # Creates repo if not exists and fetches
    return r


cli.add_command(update)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals


class VCSPullException(Exception):

    """Standard exception raised by libvcs."""

    pass


class MultipleConfigWarning(VCSPullException):
    message = 'Multiple configs found in home directory use only one.' ' .yaml, .json.'



# -*- coding: utf-8 -*-
"""Config utility functions for vcspull.
vcspull.config
~~~~~~~~~~~~~~

A lot of these items are todo.

"""
from __future__ import absolute_import, print_function, unicode_literals

import fnmatch
import glob
import logging
import os

import kaptan

from libvcs._compat import string_types

from . import exc
from .util import CONFIG_DIR, update_dict

log = logging.getLogger(__name__)


def expand_dir(_dir, cwd=os.getcwd()):
    """Return path with environmental variables and tilde ~ expanded.

    :param _dir:
    :type _dir: str
    :param cwd: current working dir (for deciphering relative _dir paths)
    :type cwd: str
    :rtype; str
    """
    _dir = os.path.expanduser(os.path.expandvars(_dir))
    if not os.path.isabs(_dir):
        _dir = os.path.normpath(os.path.join(cwd, _dir))
    return _dir


def extract_repos(config, cwd=os.getcwd()):
    """Return expanded configuration.

    end-user configuration permit inline configuration shortcuts, expand to
    identical format for parsing.

    :param config: the repo config in :py:class:`dict` format.
    :type config: dict
    :param cwd: current working dir (for deciphering relative paths)
    :type cwd: str
    :rtype: list

    """
    configs = []
    for directory, repos in config.items():
        for repo, repo_data in repos.items():

            conf = {}

            '''
            repo_name: http://myrepo.com/repo.git

            to

            repo_name: { url: 'http://myrepo.com/repo.git' }

            also assures the repo is a :py:class:`dict`.
            '''

            if isinstance(repo_data, string_types):
                conf['url'] = repo_data
            else:
                conf = update_dict(conf, repo_data)

            if 'repo' in conf:
                if 'url' not in conf:
                    conf['url'] = conf.pop('repo')
                else:
                    conf.pop('repo', None)

            '''
            ``shell_command_after``: if str, turn to list.
            '''
            if 'shell_command_after' in conf:
                if isinstance(conf['shell_command_after'], string_types):
                    conf['shell_command_after'] = [conf['shell_command_after']]

            if 'name' not in conf:
                conf['name'] = repo
            if 'parent_dir' not in conf:
                conf['parent_dir'] = expand_dir(directory, cwd)

            if 'repo_dir' not in conf:
                conf['repo_dir'] = expand_dir(
                    os.path.join(conf['parent_dir'], conf['name']), cwd
                )
            if 'remotes' in conf:
                remotes = []
                for remote_name, url in conf['remotes'].items():
                    remotes.append({'remote_name': remote_name, 'url': url})
                conf['remotes'] = sorted(
                    remotes, key=lambda x: sorted(x.get('remote_name'))
                )
            configs.append(conf)

    return configs


def find_home_config_files(filetype=['json', 'yaml']):
    """Return configs of ``.vcspull.{yaml,json}`` in user's home directory."""
    configs = []

    yaml_config = os.path.expanduser('~/.vcspull.yaml')
    has_yaml_config = os.path.exists(yaml_config)
    json_config = os.path.expanduser('~/.vcspull.json')
    has_json_config = os.path.exists(json_config)

    if not has_yaml_config and not has_json_config:
        log.debug(
            'No config file found. Create a .vcspull.yaml or .vcspull.json'
            ' in your $HOME directory. http://vcspull.git-pull.com for a'
            ' quickstart.'
        )
    else:
        if sum(filter(None, [has_json_config, has_yaml_config])) > int(1):
            raise exc.MultipleConfigWarning()
        if has_yaml_config:
            configs.append(yaml_config)
        if has_json_config:
            configs.append(json_config)

    return configs


def find_config_files(
    path=['~/.vcspull'], match=['*'], filetype=['json', 'yaml'], include_home=False
):
    """Return repos from a directory and match. Not recursive.

    :param path: list of paths to search
    :type path: list
    :param match: list of globs to search against
    :type match: list
    :param filetype: list of filetypes to search against
    :type filetype: list
    :param include_home: Include home configuration files
    :type include_home: bool
    :raises:
        - LoadConfigRepoConflict: There are two configs that have same path
          and name with different repo urls.
    :returns: list of absolute paths to config files.
    :rtype: list

    """
    configs = []

    if include_home is True:
        configs.extend(find_home_config_files())

    if isinstance(path, list):
        for p in path:
            configs.extend(find_config_files(p, match, filetype))
            return configs
    else:
        path = os.path.expanduser(path)
        if isinstance(match, list):
            for m in match:
                configs.extend(find_config_files(path, m, filetype))
        else:
            if isinstance(filetype, list):
                for f in filetype:
                    configs.extend(find_config_files(path, match, f))
            else:
                match = os.path.join(path, match)
                match += ".{filetype}".format(filetype=filetype)

                configs = glob.glob(match)

    return configs


def load_configs(files, cwd=os.getcwd()):
    """Return repos from a list of files.

    :todo: Validate scheme, check for duplciate destinations, VCS urls

    :param files: paths to config file
    :type files: list
    :param cwd: current path (pass down for :func:`extract_repos`
    :type cwd: str
    :returns: expanded config dict item
    :rtype: list of dict
    """
    repos = []
    for f in files:
        _, ext = os.path.splitext(f)
        conf = kaptan.Kaptan(handler=ext.lstrip('.')).import_config(f)

        newrepos = extract_repos(conf.export('dict'), cwd)

        if not repos:
            repos.extend(newrepos)
            continue

        dupes = detect_duplicate_repos(repos, newrepos)

        if dupes:
            msg = ('repos with same path + different VCS detected!', dupes)
            raise exc.VCSPullException(msg)
        repos.extend(newrepos)

    return repos


def detect_duplicate_repos(repos1, repos2):
    """Return duplicate repos dict if repo_dir same and vcs different.

    :param repos1: list of repo expanded dicts
    :type repos1: list of :py:dict
    :param repos2: list of repo expanded dicts
    :type repos2: list of :py:dict
    :rtype: list of dicts or None
    :returns: Duplicate lists
    """
    dupes = []
    path_dupe_repos = []

    curpaths = [r['repo_dir'] for r in repos1]
    newpaths = [r['repo_dir'] for r in repos2]
    path_duplicates = list(set(curpaths).intersection(newpaths))

    if not path_duplicates:
        return None

    path_dupe_repos.extend(
        [r for r in repos2 if any(r['repo_dir'] == p for p in path_duplicates)]
    )

    if not path_dupe_repos:
        return None

    for n in path_dupe_repos:
        currepo = next((r for r in repos1 if r['repo_dir'] == n['repo_dir']), None)
        if n['url'] != currepo['url']:
            dupes += (n, currepo)
    return dupes


def in_dir(config_dir=CONFIG_DIR, extensions=['.yml', '.yaml', '.json']):
    """Return a list of configs in ``config_dir``.

    :param config_dir: directory to search
    :type config_dir: str
    :param extensions: filetypes to check (e.g. ``['.yaml', '.json']``).
    :type extensions: list
    :rtype: list

    """
    configs = []

    for filename in os.listdir(config_dir):
        if is_config_file(filename, extensions) and not filename.startswith('.'):
            configs.append(filename)

    return configs


def filter_repos(config, repo_dir=None, vcs_url=None, name=None):
    """Return a :py:obj:`list` list of repos from (expanded) config file.

    repo_dir, vcs_url and name all support fnmatch.

    :param config: the expanded repo config in :py:class:`dict` format.
    :type config: dict
    :param repo_dir: directory of checkout location, fnmatch pattern supported
    :type repo_dir: str or None
    :param vcs_url: url of vcs remote, fn match pattern supported
    :type vcs_url: str or None
    :param name: project name, fnmatch pattern supported
    :type name: str or None
    :rtype: list

    """
    repo_list = []

    if repo_dir:
        repo_list.extend(
            [r for r in config if fnmatch.fnmatch(r['parent_dir'], repo_dir)]
        )

    if vcs_url:
        repo_list.extend(
            r for r in config if fnmatch.fnmatch(r.get('url', r.get('repo')), vcs_url)
        )

    if name:
        repo_list.extend([r for r in config if fnmatch.fnmatch(r.get('name'), name)])

    return repo_list


def is_config_file(filename, extensions=['.yml', '.yaml', '.json']):
    """Return True if file has a valid config file type.

    :param filename: filename to check (e.g. ``mysession.json``).
    :type filename: str
    :param extensions: filetypes to check (e.g. ``['.yaml', '.json']``).
    :type extensions: list or str
    :rtype: bool

    """
    extensions = [extensions] if isinstance(extensions, string_types) else extensions
    return any(filename.endswith(e) for e in extensions)



__title__ = 'vcspull'
__package_name__ = 'vcspull'
__description__ = 'synchronize your repos and forks via declarative yaml/json file(s)'
__version__ = '1.3.1'
__author__ = 'Tony Narlock'
__github__ = 'https://github.com/vcs-python/vcspull'
__docs__ = 'https://vcspull.git-pull.com'
__tracker__ = 'https://github.com/vcs-python/vcspull/issues'
__pypi__ = 'https://pypi.org/project/vcspull/'
__email__ = 'tony@git-pull.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2013-Tony Narlock'
